@@ -1,3 +1,5 @@
+use std::fmt;
+
 use crate::{
     error::{cuda_error, CudaResult},
     sys, Cuda,
@@ -19,6 +21,27 @@ impl From<u32> for CudaVersion {
     }
 }
 
+impl Into<(u32, u32)> for CudaVersion {
+    fn into(self) -> (u32, u32) {
+        (self.major, self.minor)
+    }
+}
+
+impl From<(u32, u32)> for CudaVersion {
+    fn from(other: (u32, u32)) -> Self {
+        CudaVersion {
+            major: other.0,
+            minor: other.1,
+        }
+    }
+}
+
+impl fmt::Display for CudaVersion {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "{}.{}", self.major, self.minor)
+    }
+}
+
 impl Cuda {
     /// Gets the local driver version (not to be confused with device compute capability)
     pub fn version() -> CudaResult<CudaVersion> {
